//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "weircd.h"
 
void* examinemodule;
int unloadexaminemodule;
int servicesmodule_enabledline = 0;

static int CheckSpace(const char* val) {
	if (charpos(val,' ') >= 0) {
		cerr("Invalid value - contains space chars",creader_GetLine());
		return -1;
	}
	return 0;
}
static int CheckNumber(const char* val) {
	if (issimplenumber(val) < 0) {
		cerr("Invalid value - not a number",creader_GetLine());
		return -1;
	}
	if (atoi(val) <= 0) {
		cerr("Invalid value - must be greater than zero",creader_GetLine());
		return -1;
	}
	return 0;
}
static int SetInt(const char* tg, const char* val) {
	if (CheckNumber(val) < 0) {return -1;}
	if (config_IsValSet(tg) == 0) {
		cerr("Value already defined!",creader_GetLine());return -1;
	}
	config_SetIntVal(tg,atoi(val));
	return 0;
}
static int SetInt2(const char* tg, int val) {
	if (config_IsValSet(tg) == 0) {
		cerr("Value already defined!",creader_GetLine());return -1;
	}
	config_SetIntVal(tg,val);
	return 0;
}
static int SetStr(const char* tg, const char* val) {
	if (config_IsValSet(tg) == 0) {
		cerr("Value already defined!",creader_GetLine());return -1;
	}
	config_SetStrVal(tg,val);
	return 0;
}
static int ircop_ParseFlag(const char* opname, const char* flag) {
	if (strlen(flag) <= 0) {
		cerr("Keyword expected",creader_GetLine());
		return -1;
	}
	if (ircopinfo_SetFlag(opname,flag) < 0) {
		char errmsg[300];
		snprintf(errmsg,299,"Unknown oper flag '%s'",flag);errmsg[299] = 0;
		cerr(errmsg,creader_GetLine());
		return -1;
	}
	return 0;
}
int clineparser_Do(const char* mainblock, const char* subblock, const char* linecontent) {
	char* args = alloca(strlen(linecontent)+1);
	strcpy(args,"");
	char* line = alloca(strlen(linecontent)+1);
	int linenumber = creader_GetLine();
	strcpy(line,linecontent);
	if (charpos(line,'=') > 0) {
		strcpy(args,line);
		line[charpos(line,'=')] = 0;
		trimright(line);
		cut(args,charpos(args,'=')+1);trimleft(args);
		if (strlen(args) <= 0) {
			cerr("Expecting value after '='",linenumber);return -1;
		}
	}
	//printf("parsing line -> %s = %s\n",line,args);
	if (strcasecmp(mainblock,"listen") == 0) {
		if (strcasecmp(line,"bind") == 0) {
			char ip[IPMAXLEN+5];
			char port[10];
			strcopy(ip,args,IPMAXLEN+4);
			if (charpos(ip,'/') <= 0) {
				if (charpos(ip,'/') == 0) {
					cerr("Expecting value in front of '/' seperator",linenumber);return -1;
				}
				cerr("Expecting '/' seperator",linenumber);return -1;
			}
			cut(args,charpos(ip,'/')+1);trimleft(args);
			ip[charpos(ip,'/')] = 0;trimright(ip);
			strcopy(port,args,9);
			if (charpos(port,'/') <= 0) {
				if (charpos(port,'/') == 0) {
					cerr("Expecting value in front of '/' seperator",linenumber);return -1;
				}
				cerr("Expecting '/' seperator",linenumber);return -1;
			}
			cut(args,charpos(args,'/')+1);
			port[charpos(port,'/')] = 0;
			if (issimplenumber(port) != 0 || atoi(port) <= 0 || atoi(port) > 99999) {
				cerr("Expecting 1-99999 number for second parameter",linenumber);return -1;
			}
			if (strcasecmp(args,"ipv4") != 0 && strcasecmp(args,"ipv6") != 0) {
				cerr("Expecting 'ipv4' or 'ipv6' for third parameter",linenumber);return -1;
			}
			if (strcasecmp(args,"ipv4") == 0 && isipv4ip(ip) != 0 && strcasecmp(ip,"any") != 0) {
				cerr("Expecting valid ipv4 address or 'any' as first parameter",linenumber);return -1;
			}
			if (strcasecmp(args,"ipv6") == 0 && isipv6ip(ip) != 0 && strcasecmp(ip,"any") != 0) {
				cerr("Expecting valid ipv6 address or 'any' as first parameter",linenumber);return -1;
			}
			//printf(">> bind %s:%d (%s)\n",ip,atoi(port),args);
			int iptype = IPTYPE_IPV4;
			if (strcasecmp(args,"ipv6") == 0) {
				#ifndef IPV6
				cerr("This version of WeIRCd is compiled without IPv6 support",linenumber);return -1;
				#endif
				iptype = IPTYPE_IPV6;
			}
			int r = conflisteners_Add(atoi(port),ip,iptype);
			if (r < 0) {
				if (r == -2) {
					cerr("Double definition of ports",linenumber);
				}
				if (r == -1) {
					cerr(error_mem,linenumber);
				}
				return -1;
			}
			return 0;
		}
	}
	if (strcasecmp(mainblock,"server") == 0) {
		if (strcasecmp(line,"name") == 0) {
			if (strlen(args) > 50) {
				return -1;
			}
			if (charpos(args,'!') >= 0) {
				cerr("Invalid value - contains forbidden character '!'",linenumber);return -1;
			}
			if (charpos(args,'@') >= 0) {
				cerr("Invalid value - contains forbidden character '@'",linenumber);return -1;
			}
			if (CheckSpace(args) < 0) {return -1;}
			if (SetStr("servername",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"description") == 0) {
			if (strlen(args) > 50) {
				cerr("Invalid value - too long (max 50 chars)",linenumber);return -1;
			}
			if (strlen(args) < 2) {
				cerr("Invalid value - too short (minimum 2 chars)",linenumber);return -1;
			}
			if (SetStr("serverdesc",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"password") == 0) {
			if (strlen(args) > 50) {
				cerr("Invalid value - too long (max 50 chars)",linenumber);return -1;
			}
			if (strlen(args) < 2) {
				cerr("Invalid value - too short (minimum 2 chars)",linenumber);return -1;
			}
			if (CheckSpace(args) < 0) {return -1;}
			if (SetStr("pass",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"motdfile") == 0) {
			if (SetStr("motdfilepath",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"glinefile") == 0) {
			if (SetStr("glinefilepath",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"klinefile") == 0) {
			if (SetStr("klinefilepath",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"helpchan") == 0) {
			if (args[0] != '#') {
				cerr("Invalid value - not a channel name",linenumber);return -1;
			}
			if (CheckSpace(args) < 0) {return -1;}
			if (SetStr("helpchan",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"maxchansperuser") == 0) {
			if (SetInt("maxchansperuser",args) < 0) {return -1;}
			return 0;
		}
		if (strcasecmp(line,"network") == 0) {
			if (strlen(args) > 30) {
				cerr("Invalid value - too long (max 30 chars)",linenumber);return -1;
			}
			if (CheckSpace(args) < 0) {return -1;}
			if (SetStr("network",args) < 0) {return -1;}
			return 0;
		}
	}
	if (strcasecmp(mainblock,"Services") == 0) {
		if (subblock != NULL) {
			if (servicesmodule_enabledline == 0) {
				if (strcasecmp(line,"Enabled") != 0) {
					cerr("Invalid value - Please start with the Enabled option in services blocks",linenumber);return -1;
				}else{
					if (strcasecmp(args,"false") == 0) {
						servicesmodule_enabledline = 2;
						return 0;
					}
					if (strcasecmp(args,"true") == 0) {
						servicesmodule_enabledline = 1;
						//get ready for config checking
						modlist_AddAsServices(subblock);
						servicesmodule_InitConfigPreCheck(subblock);
						return 1;
					}
					cerr("Invalid value - expected 'true' or 'false'",linenumber);
					return -1;
				}
			}
			if (servicesmodule_enabledline == 2) {return 0;}//skip all options
			if (strcasecmp(line,"Enabled") == 0) {
				cerr("Enabled directive already used",linenumber);return -1;
			}
			char msg[500];
			if (servicesmodule_ConfigPreCheckOption(subblock,line,args) != 0) {
				snprintf(msg,499,"Option '%s' is invalid or has invalid value",line);msg[499] = 0;
				cerr(msg,linenumber);
				return -1;
			}
			modlist_AddServicesOption(subblock,line,args);
			return 0;
		}else{

		}
	}
	if (strcasecmp(mainblock,"WebIRC") == 0) {
		if (strcasecmp(line,"Type") == 0) {
			if (strlen(args) < 2) {
				cerr("Invalid value - too short (min 2 chars)",linenumber);return -1;
			}
			if (webirc_SetType(args) != 0) {
				cerr("Value already defined!",creader_GetLine());return -1;
			}
			return 0;
		}
		if (strcasecmp(line,"Host") == 0) {
			if (strlen(args) < 2) {
				cerr("Invalid value - too short (min 2 chars)",linenumber);return -1;
			}
			if (webirc_SetHost(args) != 0) {
				cerr("Value already defined!",creader_GetLine());return -1;
			}
			return 0;
		}
		if (strcasecmp(line,"Password") == 0) {
			if (strlen(args) < 2) {
				cerr("Invalid value - too short (min 2 chars)",linenumber);return -1;
			}
			if (webirc_SetPassword(args) != 0) {
				cerr("Value already defined!",creader_GetLine());return -1;
			}
			return 0;
		}
		if (strcasecmp(line,"Ident") == 0) {
			if (strlen(args) < 2) {
				cerr("Invalid value - too short (min 2 chars)",linenumber);return -1;
			}
			if (webirc_SetIdent(args) != 0) {
				cerr("Value already defined!",creader_GetLine());return -1;
			}
			return 0;
		}
	}
	if (strcasecmp(mainblock,"IrcOps") == 0) {
		if (subblock != NULL) {
			if (strcasecmp(line,"Password") == 0) {
				if (strlen(args) < 2) {
					cerr("Invalid value - too short (min 2 chars)",linenumber);return -1;
				}
				if (CheckSpace(args) < 0) {return -1;}
				if (strcasecmp(args,"XXXX") == 0) {
					cerr("Do NEVER use the default password 'XXXX'!!",linenumber);return -1;
				}
				int r = ircopinfo_SetPass(subblock,args);
				if (r < 0) {
					if (r == -1) {
						cerr(error_mem,linenumber);
					}
					if (r == -2) {
						cerr("Password already specified",linenumber);
					}
					return -1;
				}
				return 0;
			}
			if (strcasecmp(line,"Host") == 0) {
				if (strlen(args) < 2) {
					cerr("Invalid value - too short (min 2 chars)",linenumber);return -1;
				}
				if (CheckSpace(args) < 0) {return -1;}
				int r = ircopinfo_SetHost(subblock,args);
				if (r < 0) {
					if (r == -1) {
						cerr(error_mem,linenumber);
					}
					if (r == -2) {
						cerr("Host already specified",linenumber);
					}
					return -1;
				}
				return 0;
			}
			if (strcasecmp(line,"Flags") == 0) {
				while (charpos(args,'+') >= 0) {
					strcpy(line,args);
					line[charpos(line,'+')] = 0;
					cut(args,charpos(args,'+')+1);
					trimleft(line);trimright(line);
					if (ircop_ParseFlag(subblock,line) < 0) {
						return -1;
					}
				}
				trimleft(args);trimright(args);
				if (ircop_ParseFlag(subblock,args) < 0) {
					return -1;
				}
				return 0;
			}
		}
	}
	if (strcasecmp(mainblock,"NoOpChans") == 0) {
		if (strlen(line) > 0) {
			if (CheckSpace(line) < 0) {return -1;}
			if (noopchan_Add(line) < 0) {
				cerr(error_mem,linenumber);return -1;
			}
		}
		return 0;
	}
	if (strcasecmp(mainblock,"FloodProtect") == 0) {
		if (subblock != NULL) {
			if (strcasecmp(subblock,"HostCloaking") == 0) {
				if (strcasecmp(line,"Unregistered") == 0) {
					if (strlen(args) > 50) {
						cerr("Invalid value - too long (max 50 chars)",linenumber);return -1;
					}
					if (CheckSpace(args) < 0) {return -1;}
					if (SetStr("vhost_guest",args) < 0) {return -1;}
					return 0;
				}
				if (strcasecmp(line,"Registered") == 0) {
					if (strlen(args) > 50) {
						cerr("Invalid value - too long (max 50 chars)",linenumber);return -1;
					}
					if (CheckSpace(args) < 0) {return -1;}
					if (SetStr("vhost_user",args) < 0) {return -1;}
					return 0;
				}
			}
			if (strcasecmp(subblock,"ReservedNicks") == 0) {
				if (strlen(line) > 0) {
					if (CheckSpace(line) < 0) {return -1;}
					if (reservednicklist_Add(line) < 0) {
						cerr("Failed to allocate memory",linenumber);return -1;
					}
				}
				return 0;
			}
		}else{
			if (strcasecmp(line,"Level") == 0) {
				if (strcasecmp(args,"strict") != 0 && strcasecmp(args,"medium") != 0 && strcasecmp(args,"soft") != 0) {
					cerr("Invalid value - possible values are: strict, medium, soft",linenumber);return -1;
				}
				int i = FLOODLEVEL_STRICT;
				if (strcasecmp(args,"medium") == 0) {i = FLOODLEVEL_MEDIUM;}
				if (strcasecmp(args,"soft") == 0) {i = FLOODLEVEL_SOFT;}
				if (SetInt2("floodlevel",i) < 0) {return -1;}
				return 0;
			}
			if (strcasecmp(line,"MaxConnsPerIp") == 0) {
				if (SetInt("maxconnsperip",args) < 0) {return -1;}
				return 0;
			}
		}
	}
	char msg[200];
	strcpy(msg,"Unknown config option '");
	stradd(msg,line,199);msg[199] = 0;
	stradd(msg,"' in block '",199);msg[199] = 0;
	stradd(msg,mainblock,199);msg[199] = 0;
	stradd(msg,"'",199);msg[199] = 0;
	if (subblock != NULL) {
		stradd(msg," in sub block '",199);msg[199] = 0;
		stradd(msg,subblock,199);msg[199] = 0;
		stradd(msg,"'",199);msg[199] = 0;
		
	}
	cerr(msg,linenumber);
	return -1;
}