//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//managing the config vars here (saved config options)

#include "weircd.h"

struct confvar {
	int defaultval;//only used for integers
	int type;
	int optional; //if set to 1 the value is not explicitely needed
	void* newdata;
	void* data;
	char* name;
};

#define MAXCONFVARS 500
struct confvar* config_confvars[MAXCONFVARS];
int confvarsinited = 0;

int lastchangedvar = -1;

void configvars_Init() {
	if (confvarsinited == 1) {return;}
	int r = 0;
	while (r < MAXCONFVARS) {
		config_confvars[r] = NULL;
		r++;
	}
	confvarsinited = 1;
}
const char config_strval_fallback[] = "";
const char* config_StrVal(const char* name) {
	configvars_Init();
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (config_confvars[r]->type == CONFVAR_STRING) {
				if (strcasecmp(config_confvars[r]->name,name) == 0) {
					if (config_confvars[r]->data != NULL) {
						return config_confvars[r]->data;
					}else{
						return config_strval_fallback;
					}
				}
			}
		}
		r++;
	}
	return config_strval_fallback;
}
int config_IntVal(const char* name) {
	configvars_Init();
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (config_confvars[r]->type == CONFVAR_INT) {
				if (strcasecmp(config_confvars[r]->name,name) == 0) {
					if (config_confvars[r]->data != NULL) {
						return *(int*)config_confvars[r]->data;
					}else{
						return 0;
					}
				}
			}
		}
		r++;
	}
	return 0;
}
const char* svname() {
	return config_StrVal("servername");
}
int config_IsValSet(const char* name) {
	configvars_Init();
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (strcasecmp(config_confvars[r]->name,name) == 0) {
				if (config_confvars[r]->newdata != NULL) {
					return 0;
				}
				return -1;
			}
		}
		r++;
	}
	return -1;
}
int config_SetStrVal(const char* name, const char* content) {
	configvars_Init();
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (config_confvars[r]->type == CONFVAR_STRING) {
				if (strcasecmp(config_confvars[r]->name,name) == 0) {
					if (config_confvars[r]->newdata != NULL) {
						free(config_confvars[r]->newdata);
					}
					config_confvars[r]->newdata = malloc(strlen(content)+1);
					if (config_confvars[r]->newdata == NULL) {
						return -1;
					}
					strcpy(config_confvars[r]->newdata,content);
					lastchangedvar = r;
					//printf("set %s to %s\n",config_confvars[r]->name,config_confvars[r]->newdata);
					return 0;
				}
			}
		}
		r++;
	}
	return -1;
}
int config_SetIntVal(const char* name, int val) {
	configvars_Init();
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (config_confvars[r]->type == CONFVAR_INT) {
				if (strcasecmp(config_confvars[r]->name,name) == 0) {
					if (config_confvars[r]->newdata != NULL) {
						*(int*)config_confvars[r]->newdata = val;
						return 0;
					}
					config_confvars[r]->newdata = malloc(sizeof(int));
					if (config_confvars[r]->newdata == NULL) {
						return -1;
					}
					*(int*)config_confvars[r]->newdata = val;
					lastchangedvar = r;
					return 0;
				}
			}
		}
		r++;
	}
	return -1;
}
void config_FreeConfVar(int r) {
	if (config_confvars[r]->name != NULL) {
		free(config_confvars[r]->name);
	}
	if (config_confvars[r]->data != NULL) {
		free(config_confvars[r]->data);
	}
	if (config_confvars[r]->newdata != NULL) {
		free(config_confvars[r]->newdata);
	}
	free(config_confvars[r]);
}
int configvars_AddTemplate(int type, const char* name) {
	configvars_Init();
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] == NULL) {
			config_confvars[r] = malloc(sizeof(struct confvar));
			if (config_confvars[r] == NULL) {
				return -1;
			}
			memset(config_confvars[r],0,sizeof(struct confvar));
			config_confvars[r]->name = malloc(strlen(name)+1);
			if (config_confvars[r]->name == NULL) {
				config_FreeConfVar(r);
				config_confvars[r] = NULL;
				return -1;
			}
			strcpy(config_confvars[r]->name,name);
			config_confvars[r]->type = type;
			lastchangedvar = r;
			return r;
		}
		r++;
	}
	return -1;
}
void configvars_ClearTemplates() {
	configvars_Init();
	lastchangedvar = -1;
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			config_FreeConfVar(r);
			config_confvars[r] = NULL;
		}
		r++;
	}
}
void configvars_Confirm() {
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (config_confvars[r]->data != NULL) {
				free(config_confvars[r]->data);
			}
			config_confvars[r]->data = config_confvars[r]->newdata;
			config_confvars[r]->newdata = NULL;
			/*if (config_confvars[r]->type == CONFVAR_INT) {
				config_confvars[r]->newdata = malloc(sizeof(int));
				if (config_confvars[r]->newdata != NULL) {
					*(int*)config_confvars[r]->newdata = config_confvars[r]->defaultvar;
				}
			}*/
		}
		r++;
	}
}
void configvars_SetOptional() {
	if (lastchangedvar < 0) {return;}
	config_confvars[lastchangedvar]->optional = 1;
}
const char* confvars_printname(int var) {
	const char* p = config_confvars[var]->name;
	if (strcasecmp(p,"servername") == 0) {
		return "Name";
	}
	if (strcasecmp(p,"serverdesc") == 0) {
		return "Description";
	}
	return p;
}
extern void* ircoplist_new;
int configvars_PromptMissing() {
	int r = 0;
	while (r < MAXCONFVARS) {
		if (config_confvars[r] != NULL) {
			if (config_confvars[r]->newdata == NULL) {
				if (config_confvars[r]->optional == 0) {
					char missingmsg[500];
					snprintf(missingmsg,499,"Value '%s' not specified",confvars_printname(r));
					missingmsg[499] = 0;
					cerr(missingmsg,-1);
					return -1;
				}
			}else{
				/*if (config_confvars[r]->type == CONFVAR_INT) {
					printf("'%s' is %d\n",config_confvars[r]->name,*(int*)config_confvars[r]->newdata);
				}
				if (config_confvars[r]->type == CONFVAR_STRING) {
					printf("'%s' is %s\n",config_confvars[r]->name,(const char*)config_confvars[r]->newdata);
				}*/
			}
		}
		r++;
	}
	int ircopcount = 0;
	void* item = ircoplist_new;
	char missingmsg[500];
	while (item != NULL) {
		struct ircopinfo* i = linkedlist_GetData(item);
		if (i != NULL) {
			ircopcount++;
			if (i->pass == NULL) {
				snprintf(missingmsg,499,"Password for IRC op '%s' not specified",i->name);
				missingmsg[499] = 0;
				cerr(missingmsg,-1);
				return -1;
			}
		}
		item = linkedlist_NextItem(item);
	}
	if (ircopcount <= 0) {
		cerr("No IRC op accounts specified.",-1);
		return -1;
	}
	extern void* webirclist_new;
	item = webirclist_new;
	while (item != NULL) {
		struct webirc* w = linkedlist_GetData(item);
		if (strlen(w->password) <= 0) {
			cerr("Password for WEBIRC entry not specified",-1);
			return -1;
		}
		if (strlen(w->host) <= 0) {
			cerr("Target host for WEBIRC entry not specified",-1);
			return -1;
		}
		item = linkedlist_NextItem(item);
	}
	return 0;
}