//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "weircd.h"

//-> called by masks_Match()
//-> see if a specific * finds a fitting end (also invoking further starmatch's when there are more *)
int starmatch(const char* mstring, const char* mask, int recursivecount) { 
	if (recursivecount < 0) {return -1;}
	//store string lengths
	int masklen = strlen(mask);
	int stringlen = strlen(mstring);
	if (masklen == 0) {
		return -1; //shouldn't happen
	}
	if (mask[0] == '*') {return -1;}//this shouldn't happen
	//short precheck to avoid further parsing:
	
	//ok. now we search a char in "mstring" that matches mask[1] and could therefore be the matching ending after the *
	//loop through all chars:
	int trypos = 0;
	while (trypos < stringlen) {
		if (mstring[trypos] == mask[0] || mask[0] == '?') {//ok matches. check from hereon if the rest also fits
			int r,k;
			r = trypos+1;//cycle the string from this position
			k = 1; //cycle the mask from this position
			//ok, check if the rest also matches. if not, quit the inner loop and advance in the outer one
			while (r < stringlen) {
				if (k > masklen - 1) {
					//string is longer than mask. mask didn't end with * -> no match
					//(if mask ended with * the condition below would have been triggered)
					break;
				}
				if (mask[k] == '*') {
					//ok a star. invoke another star match
					//are we at the end of the mask?
					if (k == masklen-1) {
						return 0;//we are! that means just anything is matched.
					}
					//ok invoke a starmatch check. if it succeeds then we're done
					if (starmatch(mstring+r,mask+k+1,recursivecount - 1) != 0) {
						//didn't succeed. ok, this is a dead end. quit to outer loop
						break;
					}else{
						//worked! that means this fits. return success
						return 0;
					}
				}else{
					if (mstring[r] != mask[k] && mask[k] != '?') {
						//a char does not match -> dead end. return to outer loop
						break;
					}else{
						//this char matched. are we at the end yet?
						if (k >= strlen(mask) - 1 && r >= strlen(mstring) - 1) {
							//we are at the last char. everything matched. success!
							return 0;
						}
						//are we at the end of the string but not of the mask?
						if (r >= strlen(mstring) - 1) {
							//we are. is the mask just having a * char follow?
							if (k == masklen-2) {
								if (mask[masklen-1] == '*') {
									//ok, * also matches nothing. so we are done!
									return 0;
								}else{
									//no, some real char that still needs to be matched.
									// -> dead end, quit inner loop and check for more in outer
									break;
								}
							}
						}
					}
				}
				r += 1;k += 1;
			}
		}
		trypos += 1;
	}
	return -1;//leider hat nix gepasst. damit matcht das ganze auch ned.
}
int masks_Match(const char* mstring,const char* mask) {
	//revised for weircd 0.7, parses *? placeholders (weircd 0.6 only *)
	int pos = 0;
	int r = 0;
	char* lclmask = alloca(strlen(mask)+1);//save a copy of our mask
	char* lclstring = alloca(strlen(mstring)+1);//save a copy of our string

	//get our local copies and make them lowercase so we automatically ignore case when comparing
	strcpy(lclmask,mask);strlwr(lclmask);
	strcpy(lclstring,mstring);strlwr(lclstring);
	
	//remove doubled stars
	int len = strlen(lclmask);
	while (r < len - 1) {
		if (lclmask[r] == '*' && lclmask[r + 1] == '*') {
			//ok, memmove that char away
			memmove(&lclmask[r+1],&lclmask[r+2],strlen(lclmask)-(r+1));
			len--;
		}
		r += 1;
	}

	pos = 0;
	//ok, now compare every single char and invoke recursion on * chars to parse them
	len = strlen(lclmask);
	int stringlen = strlen(lclstring);
	while (pos < len) {
		if (pos >= stringlen) {
			//we are at the end? when the mask just has a * here this is no problem
			if (lclmask[pos] == '*' && pos == len-1) {return 0;}
			return -1;
		}
		//check if current char matches
		if (lclmask[pos] != '*') {
			//just a normal char or a ?. if it doesn't match, quit
			if (lclmask[pos] != lclstring[pos] && lclmask[pos] != '?') {return -1;}
		}else{
			//it's a *. invoke starmatch to search for a fitting part to the after-* mask content
			if (pos == len-1) {
				//at the end of the mask!
				return 0;//the * obviously matches everything
			}
			if (starmatch(lclstring+pos,lclmask+pos+1,10) != 0) {return -1;}else{return 0;}
		}
		pos += 1;
	}
	return -1;
}
void masks_MakeValid(char* mymask) {
	mymask[MAXUSERMASKLEN] = 0;
	char lclmaskcpy[MAXUSERMASKLEN + 1] = "";
	int r = 0;
	int k = 0;
	while (r < strlen(mymask)) {
		if (mymask[r] > 30 && mymask[r] != '#' && mymask[r] != ' ' && mymask[r] != ',' && mymask[r] != 30 && mymask[r] < 127) {
			if (r < strlen(mymask) - 1) {
				if (mymask[r] == '*' && mymask[r + 1] == '*') {r += 1;continue;}//no doubled *
			}
			lclmaskcpy[strlen(lclmaskcpy) + 1] = '\0';
			lclmaskcpy[strlen(lclmaskcpy)] = mymask[r];
		}
		r += 1;
	}
	strlwr(lclmaskcpy);
	//strncpy(lclmaskcpy,mymask,MAXUSERMASKLEN);
	lclmaskcpy[MAXUSERMASKLEN - 1] = 0;
	if (charpos(lclmaskcpy,'!') > MAXUSERMASKLEN - 5) {lclmaskcpy[MAXUSERMASKLEN - 5] = 0;}
	if (charpos(lclmaskcpy,'!') <= 0) {
		if (charpos(lclmaskcpy,'!') == 0) {// ! is at the start
			lclmaskcpy[MAXUSERMASKLEN - 1] = 0;
			r = (strlen(lclmaskcpy) - 1) + 1;
			while (r > 0) {
				lclmaskcpy[r] = lclmaskcpy[r - 1];
				r -= 1;
			}
			lclmaskcpy[0] = '*';
			if (charpos(lclmaskcpy,'@') < 0) {
				if (lclmaskcpy[strlen(lclmaskcpy) - 1] == '!') {lclmaskcpy[MAXUSERMASKLEN - 4] = 0;strcat(lclmaskcpy,"*");}
				lclmaskcpy[MAXUSERMASKLEN - 3] = 0;strcat(lclmaskcpy,"@*");
			}else{
				if (charpos(lclmaskcpy,'@') == charpos(lclmaskcpy,'!') + 1) {//no userid, e.g. *!@
					lclmaskcpy[MAXUSERMASKLEN - 2] = 0;
					r = strlen(lclmaskcpy);k = charpos(lclmaskcpy,'@');
					lclmaskcpy[strlen(lclmaskcpy) + 1] = 0;
					while (r > k) {
						lclmaskcpy[r] = lclmaskcpy[r - 1];
						r -= 1;
					}
					lclmaskcpy[k] = '*';
				}
			}
			if (charpos(lclmaskcpy,'@') == strlen(lclmaskcpy) - 1) {//no host, e.g. *!*@
				lclmaskcpy[MAXUSERMASKLEN - 2] = 0;strcat(lclmaskcpy,"*");
			}
		}else{// ! is missing
			if (charpos(lclmaskcpy,'@') < 0) {// @ is missing too
				lclmaskcpy[MAXUSERMASKLEN - 4] = 0;
				strcat(lclmaskcpy,"!*@*");
			}else{
				if (charpos(lclmaskcpy,'@') == 0) { // @ is at the beginning
					lclmaskcpy[MAXUSERMASKLEN - 4] = 0;
					lclmaskcpy[(strlen(lclmaskcpy) - 1) + 5] = 0;
					r = (strlen(lclmaskcpy) - 1) + 4;
					while (r > 2) {
						lclmaskcpy[r] = lclmaskcpy[r - 3];
						r -= 1;
					}
					lclmaskcpy[0] = '*';lclmaskcpy[1] = '!';lclmaskcpy[2] = '*';lclmaskcpy[3] = '@'; 
				}else{ // @ is somewhere at the end
					lclmaskcpy[MAXUSERMASKLEN - 2] = 0;
					lclmaskcpy[(strlen(lclmaskcpy) - 1) + 3] = 0;
					r = (strlen(lclmaskcpy) - 1) + 2;
					while (r > 1) {
						lclmaskcpy[r] = lclmaskcpy[r - 2];
						r -= 1;
					}
					lclmaskcpy[0] = '*';lclmaskcpy[1] = '!';
				}
				if (charpos(lclmaskcpy,'@') == strlen(lclmaskcpy) - 1) {//no host, e.g. *!*@
					lclmaskcpy[MAXUSERMASKLEN - 2] = 0;strcat(lclmaskcpy,"*");
				}
			}
		}
	}else{
		if (charpos(lclmaskcpy,'@') < 0) {
			if (lclmaskcpy[strlen(lclmaskcpy)-1] == '!') {//no userid, e.g. *!
				lclmaskcpy[MAXUSERMASKLEN - 4] = 0;strcat(lclmaskcpy,"*");
			}
			lclmaskcpy[MAXUSERMASKLEN - 3] = 0;strcat(lclmaskcpy,"@*");
		}else{
			if (charpos(lclmaskcpy,'@') < charpos(lclmaskcpy,'!')) {
				lclmaskcpy[charpos(lclmaskcpy,'!')] = '\0';
				lclmaskcpy[MAXUSERMASKLEN - 2] = 0;
				lclmaskcpy[(strlen(lclmaskcpy) - 1) + 3] = 0;
				r = (strlen(lclmaskcpy) - 1) + 2;
				while (r > 1) {
					lclmaskcpy[r] = lclmaskcpy[r - 2];
					r -= 1;
				}
				lclmaskcpy[0] = '*';lclmaskcpy[1] = '!';
			}
			if (charpos(lclmaskcpy,'@') == charpos(lclmaskcpy,'!') + 1) {//no userid, e.g. *!@
				lclmaskcpy[MAXUSERMASKLEN - 2] = 0;
				r = strlen(lclmaskcpy);k = charpos(lclmaskcpy,'@');
				lclmaskcpy[strlen(lclmaskcpy) + 1] = 0;
				while (r > k) {
					lclmaskcpy[r] = lclmaskcpy[r - 1];
					r -= 1;
				}
				lclmaskcpy[k] = '*';
			}
			if (charpos(lclmaskcpy,'@') == strlen(lclmaskcpy) - 1) {//no host, e.g. *!*@
				lclmaskcpy[MAXUSERMASKLEN - 2] = 0;strcat(lclmaskcpy,"*");
			}
		}
	}
	strcpy(mymask,lclmaskcpy);
}

//a small testing function to print out masks_Match() results:
void masks_Test(const char* string, const char* mask, int expectedresult) {
	int i = masks_Match(string,mask);
	if (i != expectedresult) {
		printf("FAIL: ");
	}else{
		printf(" OK : ");
	}
	printf(" does %s match %s? result: ",mask,string);
	if (i == 0) {
		printf("yes");
	}else{
		printf("no");
	}
	printf("\n");
}